import logging
import os
import pkgutil
import time
import traceback
import argparse
import pickle
from sklearn.model_selection import StratifiedShuffleSplit, StratifiedKFold


import matplotlib
import numpy as np

# Import own modules
from . import monoview_classifiers
from . import multiview_classifiers
from .monoview.exec_classif_mono_view import MonoViewExp
from .multiview.exec_multiview import MultiViewExp
from .result_analysis.execution import analyze_iterations, analyze
from .utils import execution, dataset, configuration
from .utils.execution import BaseExec
from .utils.configuration import save_config
from .utils.dataset import delete_HDF5
from .utils.organization import secure_file_path
from .utils import get_multiview_db as DB


matplotlib.use(
    'Agg')  # Anti-Grain Geometry C++ library to make a raster (pixel) image of the figure

# Author-Info
__author__ = "Baptiste Bauvin"
__status__ = "Prototype"  # Production, Development, Prototype

package_path = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

class BootstrapIteration:

    def __init__(self,
                labels_dictionary={},
                directory="dir",
                classification_indices=[],
                args={},
                k_folds=StratifiedKFold(),
                random_state=42,
                hyper_param_search="Random",
                metrics={},
                argument_dictionaries=[],
                monoview_benchmark=[],
                multiview_benchmark=[],
                views=[],
                views_indices=[],
                flag=1,track_tracebacks=True,
                database_name=""):
        self.labels_dictionary=labels_dictionary
        self.directory = directory
        self.classification_indices = classification_indices
        self.args = args
        self.k_folds = k_folds
        self.random_state = random_state
        self.hyper_param_search = hyper_param_search
        self.metrics = metrics
        self.argument_dictionaries = argument_dictionaries
        self.monoview_benchmark = monoview_benchmark
        self.multiview_benchmark = multiview_benchmark
        self.views = views
        self.views_indices = views_indices
        self.flag = flag
        self.track_tracebacks = track_tracebacks
        self.traceback_outputs = {}
        self.results = []
        self.database_name=database_name
        self.test_results = []

    def benchmark_init(self, dataset_var):
        """
        Initializes the benchmark, by saving the indices of the train
        samples and the cross validation folds.

        Parameters
        ----------
        directory : str
            The benchmark's result directory

        classification_indices : numpy array
            The indices of the samples, splitted for the train/test split

        labels : numpy array
            The labels of the dataset

        labels_dictionary : dict
            The dictionary with labels as keys and their names as values

        k_folds : sklearn.model_selection.Folds object
            The folds for the cross validation process

        Returns
        -------

        """
        logging.info("Start:\t Benchmark initialization")
        secure_file_path(os.path.join(self.directory, "train_labels.csv"))
        train_indices = self.classification_indices[0]
        train_labels = dataset_var.get_labels(sample_indices=train_indices)
        np.savetxt(os.path.join(self.directory, "train_labels.csv"), train_labels,
                   delimiter=",")
        np.savetxt(os.path.join(self.directory, "train_indices.csv"),
                   self.classification_indices[0],
                   delimiter=",")
        self.results_monoview = []
        folds = self.k_folds.split(np.arange(len(train_labels)), train_labels)
        min_fold_len = int(len(train_labels) / self.k_folds.n_splits)
        for fold_index, (train_cv_indices, test_cv_indices) in enumerate(folds):
            file_name = os.path.join(self.directory, "folds", "test_labels_fold_" + str(
                fold_index) + ".csv")
            secure_file_path(file_name)
            np.savetxt(file_name, train_labels[test_cv_indices[:min_fold_len]],
                       delimiter=",")
        self.labels_names = list(self.labels_dictionary.values())
        logging.info("Done:\t Benchmark initialization")

    def exec_one_benchmark_mono_core(self, dataset_var):  # pragma: no cover
        self.benchmark_init(dataset_var)
        logging.getLogger('matplotlib.font_manager').disabled = True
        logging.info("Start:\t Test Benchmark")

        for exp_ind, exp in enumerate(self.argument_dictionaries):
            try:
                exp.add_bootstrap_info(self.directory, self.k_folds,
                                       self.classification_indices,
                                       self.random_state)
                self.results.append(exp.exec(dataset_var) )
                self.argument_dictionaries[exp_ind]=exp
            except BaseException:
                if self.track_tracebacks:
                    if isinstance(exp, MonoViewExp):
                        self.traceback_outputs[
                            exp.classifier_name + "-" + exp.view_name] = traceback.format_exc()
                    else:
                        self.traceback_outputs[exp.classifier_name] = traceback.format_exc()
                else:
                    raise
        logging.info("Done:\tTest benchmark")

    def exec_one_test_benchmark(self, dataset_var, y=None, feature_ids=None,
                                n_splits=5, test_size=0.1):
        logging.getLogger('matplotlib.font_manager').disabled = True
        logging.info("Start:\t Benchmark")
        self.results=[]
        for exp in self.argument_dictionaries:
            try:
                exp.add_bootstrap_info(self.directory, self.k_folds,
                                       self.classification_indices,
                                       self.random_state)
                self.results.append(exp.test(dataset_var, y, feature_ids,
                                             n_splits=n_splits,
                                             test_size=test_size))
            except BaseException:
                if self.track_tracebacks:
                    if isinstance(exp, MonoViewExp):
                        self.traceback_outputs[
                            exp.classifier_name + "-" + exp.view_name] = traceback.format_exc()
                    else:
                        self.traceback_outputs[
                            exp.classifier_name] = traceback.format_exc()
                else:
                    raise
        logging.info("Done:\t monoview benchmark")

        # logging.info("Start:\t multiview benchmark")
        # self.results_multiview = []
        # for arguments in self.argument_dictionaries["multiview"]:
        #     try:
        #         self.results_multiview += [
        #             exec_multiview(self.directory, self.dataset_var, self.name,
        #                            self.classification_indices,
        #                            self.k_folds, self.nb_cores, self.args["file_type"],
        #                            self.pathf, self.labels_dictionary, self.random_state,
        #                            self.labels,
        #                            hps_method=self.hyper_param_search,
        #                            metrics=self.metrics, n_iter=self.hps_iter,
        #                            **self.arguments)]
        #     except BaseException:
        #         if self.track_tracebacks:
        #             traceback_outputs[
        #                 arguments["classifier_name"]] = traceback.format_exc()
        #         else:
        #             raise
        logging.info("Done:\t multiview benchmark")


class Summit(BaseExec):

    def __init__(self, log=True,
                        name=["plausible", ],
                        label="_",
                        file_type=".hdf5",
                        views=None,
                        pathf=os.path.join(os.path.dirname(package_path), "data", ""),
                        nice=0,
                        random_state=42,
                        nb_cores=1,
                        full=True,
                        debug=False,
                        add_noise=False,
                        noise_std=0.0,
                        res_dir=os.path.join(os.path.dirname(package_path),"results", ""),
                        track_tracebacks=True,
                        split=0.49,
                        nb_folds=5,
                        nb_class=None,
                        classes=None,
                        cl_type=["multiview", ],
                        algos_monoview=["all"],
                        algos_multiview=["svm_jumbo_fusion", ],
                        stats_iter=2,
                        metrics={"accuracy_score": {}, "f1_score": {}},
                        metric_princ="accuracy_score",
                        hps_type="Random",
                        hps_iter=1,
                        hps_kwargs={'n_iter': 10, "equivalent_draws": True},
                        config_path=None,
                        **kwargs):
        if config_path is not None:
            if type(config_path)==list:
                config_path = self.parse_the_args(config_path).config_path
            args = configuration.get_the_args(config_path)
            Summit.__init__(self, **args)
        else:
            self.log = log
            if type(name)==list:
                self.name = name[0]
            else:
                self.name =  name
            self.label = label
            self.file_type = file_type
            self.views = views
            self.pathf = pathf
            self.nice = nice
            self.random_state = random_state
            self.nb_cores = nb_cores
            self.full = full
            self.debug = True
            self.add_noise = add_noise
            self.noise_std = noise_std
            self.res_dir = res_dir
            self.track_tracebacks = track_tracebacks
            self.split = split
            self.nb_folds = nb_folds
            self.nb_class = nb_class
            self.classes = classes
            self.cl_type = cl_type
            self.algos_monoview = algos_monoview
            self.algos_multiview = algos_multiview
            self.stats_iter = stats_iter
            self.metrics = metrics
            self.metric_princ = metric_princ
            self.hps_type = hps_type
            self.hps_iter = hps_iter
            self.hps_kwargs = hps_kwargs
            self.args = kwargs
            self.monoview_benchmark = None
            self.multiview_benchmark = None
            self.monoview_kwargs = None
            self.multiview_kwargs = None
            self.exps = []

    def exec_classif(self, ):  # pragma: no cover
        """
        Runs the benchmark with the given arguments

        Parameters
        ----------
        arguments :

        Returns
        -------


        >>> exec_classif([--config_path, /path/to/config/files/])
        >>>
        """
        start = time.time()

        import sys
        if not sys.platform in ["win32", "cygwin"]:
            os.nice(self.nice)
        if self.nb_cores == 1:
            os.environ['OPENBLAS_NUM_THREADS'] = '1'
        self.find_dataset_names()
        # noise_results = []
        # for noise_std in args["noise_std"]:

        self.init_log_file()
        self.init_random_state()

        self.init_stats_iter_random_states()

        get_database = self.get_database_function()

        self.dataset_var, self.labels_dictionary, self.name = get_database(
            self.views,
            self.pathf, self.name,
            self.nb_class,
            self.classes,
            self.random_state,
            self.full,
        )
        self.gen_splits()

        self.gen_k_folds()
        self.init_views()
        self.views_dictionary = self.dataset_var.get_view_dict()
        self.nb_views = len(self.views)
        self.nb_class = self.dataset_var.get_nb_class()

        if self.metrics == "all":
            metrics_names = [name for _, name, isPackage
                             in pkgutil.iter_modules(
                    [os.path.join(os.path.dirname(
                        os.path.dirname(os.path.realpath(__file__))),
                        'metrics')]) if
                             not isPackage and name not in ["framework",
                                                            "log_loss",
                                                            "matthews_corrcoef",
                                                            "roc_auc_score"]]
            self.metrics = dict((metric_name, {})
                           for metric_name in metrics_names)
        self.arange_metrics()

        self.init_benchmark()
        self.init_kwargs_func()
        data_base_time = time.time() - start
        self.init_argument_dictionaries()
        self.gen_direcorties_names()
        self.gen_argument_dictionaries()
        self.exec_benchmark()

    def test_clf(self, dataset_var, y=None, sample_ids=[], feature_ids=[]):
        self.exec_test_benchmark(dataset_var, y=y, sample_ids=sample_ids,
                                 feature_ids=feature_ids,
                                 n_splits=self.stats_iter, test_size=self.split)

    def exec_test_benchmark(self, dataset_var, y=None, sample_ids=[], feature_ids=[], n_splits=1, test_size=0.8):  # pragma: no cover

        self.results = []
        for bootstrap_iter in self.benchmark_argument_dictionaries:
            bootstrap_iter.exec_one_test_benchmark(dataset_var, y=y,
                                                   feature_ids=feature_ids,
                                                   n_splits=n_splits,
                                                   test_size=test_size)
            analyze_iterations([bootstrap_iter], self.stats_iter,
                               self.metrics, sample_ids=sample_ids,
                               labels=y,
                               feature_ids=[feature_ids],
                               view_names=["MST"], test=True)
            self.results += [bootstrap_iter]
        logging.info("Done:\t Executing all the needed benchmarks")

        # Do everything with flagging
        logging.info("Start:\t Analyzing predictions")
        results_mean_stds = analyze(self.results, self.stats_iter,
                                    self.benchmark_argument_dictionaries,
                                    self.metrics,
                                    self.res_dir,
                                    sample_ids,
                                    y, [feature_ids],
                                    ["MST"], test=True)
        logging.info("Done:\t Analyzing predictions")
        return results_mean_stds

    def parse_the_args(self, arguments):
        """Used to parse the args entered by the user"""

        parser = argparse.ArgumentParser(
            description='This file is used to benchmark the scores fo multiple '
                        'classification algorithm on multiview data.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            fromfile_prefix_chars='@')

        groupStandard = parser.add_argument_group('Standard arguments')
        groupStandard.add_argument('--config_path', metavar='STRING',
                                   action='store',
                                   help='Path to the hdf5 dataset or database '
                                        'folder (default: %(default)s)',
                                   default='../config_files/config.yml')
        args = parser.parse_args(arguments)
        return args


    def init_random_state(self,):
        r"""
        Used to init a random state.
        If no random state is specified, it will generate a 'random' seed.
        If the `randomSateArg` is a string containing only numbers, it will be converted in
         an int to generate a seed.
        If the `randomSateArg` is a string with letters, it must be a path to a pickled random
        state file that will be loaded.
        The function will also pickle the new random state in a file tobe able to retrieve it later.
        Tested


        Parameters
        ----------
        random_state_arg : None or string
            See function description.
        directory : string
            Path to the results directory.

        Returns
        -------
        random_state : numpy.random.RandomState object
            This random state will be used all along the benchmark .
        """

        if self.random_state is None:
            self.random_state = np.random.RandomState(self.random_state)
        else:
            try:
                seed = int(self.random_state)
                self.random_state = np.random.RandomState(seed)
            except ValueError:
                file_name = self.random_state
                with open(file_name, 'rb') as handle:
                    self.random_state = pickle.load(handle)
        with open(os.path.join(self.res_dir, "random_state.pickle"), "wb") as handle:
            pickle.dump(self.random_state, handle)

    def init_stats_iter_random_states(self,):
        r"""
        Used to initialize multiple random states if needed because of multiple statistical iteration of the same benchmark

        Parameters
        ----------
        stats_iter : int
            Number of statistical iterations of the same benchmark done (with a different random state).
        random_state : numpy.random.RandomState object
            The random state of the whole experimentation, that will be used to generate the ones for each
            statistical iteration.

        Returns
        -------
        stats_iter_random_states : list of numpy.random.RandomState objects
            Multiple random states, one for each sattistical iteration of the same benchmark.
        """
        if self.stats_iter > 1:
            self.stats_iter_random_states = [
                np.random.RandomState(self.random_state.randint(5000)) for _ in
                range(self.stats_iter)]
        else:
            self.stats_iter_random_states = [self.random_state]

    def get_database_function(self,):
        r"""Used to get the right database extraction function according to the type of database and it's name

        Parameters
        ----------
        name : string
            Name of the database.
        type_var : string
            type of dataset hdf5 or csv

        Returns
        -------
        getDatabase : function
            The function that will be used to extract the database
        """
        if self.name not in ["fake", "plausible"]:
            get_database = getattr(DB, "get_classic_db_" + self.file_type[1:])
        else:
            get_database = getattr(DB, "get_" + self.name + "_db_" + self.file_type[1:])
        return get_database

    def init_log_file(self):
        r"""Used to init the directory where the preds will be stored and the log file.

        First this function will check if the result directory already exists (only one per minute is allowed).

        If the the result directory name is available, it is created, and the logfile is initiated.

        Parameters
        ----------
        name : string
            Name of the database.
        views : list of strings
            List of the view names that will be used in the benchmark.
        cl_type : list of strings
            Type of benchmark that will be made .
        log : bool
            Whether to show the log file in console or hide it.
        debug : bool
            for debug option
        label : str  for label

        result_directory : str name of the result directory

        add_noise : bool for add noise

        noise_std : level of std noise

        Returns
        -------
        results_directory : string
            Reference to the main results directory for the benchmark.
        """
        if self.views is None:
            self.views = []
        # result_directory = os.path.join(os.path.dirname(
        #     os.path.dirname(os.path.dirname(os.path.realpath(__file__)))),
        #                                 result_directory)
        if self.debug:
            self.res_dir = os.path.join(self.res_dir, self.name,
                                            "debug_started_" + time.strftime(
                                                "%Y_%m_%d-%H_%M_%S") + "_" + self.label)
        else:
            self.res_dir = os.path.join(self.res_dir, self.name,
                                            "started_" + time.strftime(
                                                "%Y_%m_%d-%H_%M") + "_" + self.label)
        log_file_name = time.strftime("%Y_%m_%d-%H_%M") + "-" + ''.join(
            self.cl_type) + "-" + "_".join(self.views) + "-" + self.name + "-LOG.log"
        if os.path.exists(self.res_dir):  # pragma: no cover
            raise NameError("The result dir already exists, wait 1 min and retry")
        log_file_path = os.path.join(self.res_dir, log_file_name)
        os.makedirs(os.path.dirname(log_file_path))
        logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s',
                            filename=log_file_path, level=logging.INFO,
                            filemode='w')
        if self.log:
            logging.getLogger().addHandler(logging.StreamHandler())
        save_config(self.res_dir, self.__dict__)

    def gen_splits(self,):
        r"""Used to _gen the train/test splits using one or multiple random states.

        Parameters
        ----------
        labels : numpy.ndarray
            Name of the database.
        split_ratio : float
            The ratio of samples between train and test set.
        stats_iter_random_states : list of numpy.random.RandomState
            The random states for each statistical iteration.

        Returns
        -------
        splits : list of lists of numpy.ndarray
            For each statistical iteration a couple of numpy.ndarrays is stored with the indices for the training set and
            the ones of the testing set.
        """
        indices = np.arange(len(self.dataset_var.get_labels()))
        self.splits = []
        for random_state in self.stats_iter_random_states:
            folds_obj = StratifiedShuffleSplit(n_splits=1,
                                                                       random_state=random_state,
                                                                       test_size=self.split)
            folds = folds_obj.split(indices, self.dataset_var.get_labels())
            for fold in folds:
                train_fold, test_fold = fold
            train_indices = indices[train_fold]
            test_indices = indices[test_fold]
            self.splits.append([train_indices, test_indices])

    def gen_k_folds(self,):
        r"""Used to generate folds indices for cross validation for each statistical iteration.

        Parameters
        ----------
        stats_iter : integer
            Number of statistical iterations of the benchmark.
        nb_folds : integer
            The number of cross-validation folds for the benchmark.
        stats_iter_random_states : list of numpy.random.RandomState
            The random states for each statistical iteration.

        Returns
        -------
        folds_list : list of list of sklearn.model_selection.StratifiedKFold
            For each statistical iteration a Kfold stratified (keeping the ratio between classes in each fold).
        """
        if self.stats_iter > 1:
            self.folds_list = []
            for random_state in self.stats_iter_random_states:
                self.folds_list.append(
                    StratifiedKFold(n_splits=self.nb_folds,
                                                            random_state=random_state,
                                                            shuffle=True))
        else:
            if isinstance(self.stats_iter_random_states, list):
                stats_iter_random_state = self.stats_iter_random_states[0]
            self.folds_list = [StratifiedKFold(n_splits=self.nb_folds,
                                                                  random_state=stats_iter_random_state,
                                                                  shuffle=True)]

    def init_views(self,):
        r"""Used to return the views names that will be used by the
        benchmark, their indices and all the views names.

        Parameters
        ----------
        dataset_var : HDF5 dataset file
            The full dataset that wil be used by the benchmark.
        arg_views : list of strings
            The views that will be used by the benchmark (arg).

        Returns
        -------
        views : list of strings
            Names of the views that will be used by the benchmark.
        view_indices : list of ints
            The list of the indices of the view that will be used in the benchmark (according to the dataset).
        all_views : list of strings
            Names of all the available views in the dataset.
        """
        nb_view = self.dataset_var.nb_view
        if self.views is not None:
            allowed_views = self.views
            self.all_views = [str(self.dataset_var.get_view_name(view_index))
                         if not isinstance(self.dataset_var.get_view_name(view_index), bytes)
                         else self.dataset_var.get_view_name(view_index).decode("utf-8")
                         for view_index in range(nb_view)]
            self.views = []
            self.views_indices = []
            for view_index in range(nb_view):
                view_name = self.dataset_var.get_view_name(view_index)
                if isinstance(view_name, bytes):
                    view_name = view_name.decode("utf-8")
                if view_name in allowed_views:
                    self.views.append(view_name)
                    self.views_indices.append(view_index)
        else:
            self.views = [str(self.dataset_var.get_view_name(view_index))
                     if not isinstance(self.dataset_var.get_view_name(view_index), bytes)
                     else self.dataset_var.get_view_name(view_index).decode("utf-8")
                     for view_index in range(nb_view)]
            self.views_indices = range(nb_view)
            self.all_views = self.views

    def gen_direcorties_names(self,):
        r"""Used to generate the different directories of each iteration if needed.

        Parameters
        ----------
        directory : string
            Path to the results directory.
        statsIter : int
            The number of statistical iterations.

        Returns
        -------
        directories : list of strings
            Paths to each statistical iterations result directory.
        """
        if self.stats_iter > 1:
            self.directories = []
            for i in range(self.stats_iter):
                self.directories.append(os.path.join(self.res_dir, "iter_" + str(i + 1)))
        else:
            self.directories = [self.res_dir]

    def find_dataset_names(self, ):
        """This function goal is to browse the dataset directory and extrats all
         the needed dataset names."""
        package_path = os.path.dirname(
            os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
        if os.path.isdir(self.pathf):
            pass
        elif os.path.isdir(os.path.join(package_path, self.pathf)):
            self.pathf = os.path.join(package_path, self.pathf)
        else:
            raise ValueError("The provided pathf does not exist ({}) SuMMIT checks "
                             "the prefix from where you are running your script ({}) "
                             "and the summit package prefix ({}). "
                             "You may want to try with an absolute path in the "
                             "config file".format(self.pathf, os.getcwd(), package_path))
        available_file_names = [file_name.strip().split(".")[0]
                                for file_name in
                                os.listdir(self.pathf)
                                if file_name.endswith(self.file_type)]
        self.dataset_list = [self.name]
        if self.dataset_list == ["all"]:
            self.dataset_list = available_file_names
        elif isinstance(self.dataset_list, str):
            self.dataset_list = [used_name for used_name in available_file_names if
                                 self.dataset_list == used_name]
        elif len(self.dataset_list) > 1:
            selected_names = [used_name for used_name in available_file_names if
                              used_name in self.dataset_list]
            if not selected_names:
                raise ValueError(
                    "None of the provided dataset names are available. Available datasets are {}".format(
                        available_file_names))
            self.dataset_list = [used_name for used_name in available_file_names if
                                 used_name in self.dataset_list]
        elif self.dataset_list[0] in available_file_names:
            pass
        else:
            raise ValueError(
                "The asked dataset ({}) is not available in {}. \n The available ones are {}".format(
                    self.dataset_list[0], self.pathf, available_file_names))

    def gen_argument_dictionaries(self,):  # pragma: no cover
        r"""Used to generate a dictionary for each benchmark.

        One for each label combination (if multiclass), for each statistical iteration, generates an dictionary with
        all necessary information to perform the benchmark

        Parameters
        ----------
        labels_dictionary : dictionary
            Dictionary mapping labels indices to labels names.
        directories : list of strings
            List of the paths to the result directories for each statistical iteration.
        multiclass_labels : list of lists of numpy.ndarray
            For each label couple, for each statistical iteration a triplet of numpy.ndarrays is stored with the
            indices for the biclass training set, the ones for the biclass testing set and the ones for the
            multiclass testing set.
        labels_combinations : list of lists of numpy.ndarray
            Each original couple of different labels.
        indices_multiclass : list of lists of numpy.ndarray
            For each combination, contains a biclass labels numpy.ndarray with the 0/1 labels of combination.
        hyper_param_search : string
            Type of hyper parameter optimization method
        args : parsed args objects
            All the args passed by the user.
        k_folds : list of list of sklearn.model_selection.StratifiedKFold
            For each statistical iteration a Kfold stratified (keeping the ratio between classes in each fold).
        stats_iter_random_states : list of numpy.random.RandomState objects
            Multiple random states, one for each sattistical iteration of the same benchmark.
        metrics : list of lists
            metrics that will be used to evaluate the algorithms performance.
        argument_dictionaries : dictionary
            Dictionary resuming all the specific arguments for the benchmark, oe dictionary for each classifier.
        benchmark : dictionary
            Dictionary resuming which mono- and multiview algorithms which will be used in the benchmark.
        nb_views : int
            THe number of views used by the benchmark.
        views : list of strings
            List of the names of the used views.
        views_indices : list of ints
            List of indices (according to the dataset) of the used views.

        Returns
        -------
        benchmarkArgumentDictionaries : list of dicts
            All the needed arguments for the benchmarks.

        """
        self.benchmark_argument_dictionaries = []
        for iter_index, iterRandomState in enumerate(self.stats_iter_random_states):
            benchmark_argument_dictionary = BootstrapIteration(
                labels_dictionary= self.labels_dictionary,
                directory=self.directories[iter_index],
                classification_indices=self.splits[iter_index],
                args=self.args,
                k_folds=self.folds_list[iter_index],
                random_state=iterRandomState,
                hyper_param_search=self.hps_type,
                metrics=self.metrics,
                argument_dictionaries=self.exps[iter_index],
                monoview_benchmark=self.monoview_benchmark,
                multiview_benchmark=self.multiview_benchmark,
                views=self.views,
                views_indices=self.views_indices,
                flag=iter_index,
                track_tracebacks=self.track_tracebacks,
            database_name=self.name)
            self.benchmark_argument_dictionaries.append(benchmark_argument_dictionary)


    def init_benchmark(self, ):
        r"""Used to create a list of all the algorithm packages names used for the benchmark.

        First this function will check if the benchmark need mono- or/and multiview
        algorithms and adds to the right
        dictionary the asked algorithms. If none is asked by the user, all will be added.

        If the keyword `"Benchmark"` is used, all mono- and multiview algorithms will be added.

        Parameters
        ----------
        cl_type : List of string
            List of types of needed benchmark
        multiview_algos : List of strings
            List of multiview algorithms needed for the benchmark
        monoview_algos : Listof strings
            List of monoview algorithms needed for the benchmark
        args : ParsedArgumentParser args
            All the input args (used to tune the algorithms)

        Returns
        -------
        benchmark : Dictionary of dictionaries
            Dictionary resuming which mono- and multiview algorithms which will be used in the benchmark.
        """
        if "monoview" in self.cl_type:
            if self.algos_monoview == ['all']:  # pragma: no cover
                self.monoview_benchmark = [name for _, name, is_package in
                                         pkgutil.iter_modules(
                                             monoview_classifiers.__path__)
                                         if not is_package]

            else:
                self.monoview_benchmark = self.algos_monoview
        if "multiview" in self.cl_type:
            if self.algos_multiview == ["all"]:  # pragma: no cover
                self.multiview_benchmark = [name for _, name, is_package in
                                          pkgutil.iter_modules(
                                              multiview_classifiers.__path__)
                                          if not is_package]
            else:
                self.multiview_benchmark = self.algos_multiview


    def init_argument_dictionaries(self, ):  # pragma: no cover
        for iter_ind in range(self.stats_iter):
            exps=[]
            if self.monoview_benchmark is not None:
                exps += self.init_monoview_exps(
                    self.monoview_benchmark,
                    self.views_dictionary,
                    self.nb_class,
                    self.monoview_kwargs, self.hps_type, self.hps_kwargs)
            if self.multiview_benchmark is not None:
                exps += self.init_multiview_exps(
                    self.multiview_benchmark,
                    self.views_dictionary,
                    self.nb_class,
                    self.multiview_kwargs, self.hps_type, self.hps_kwargs)
            self.exps.append(exps)


    def init_multiview_exps(self, classifier_names, views_dictionary, nb_class,
                            kwargs_init, hps_method,
                            hps_kwargs):  # pragma: no cover
        multiview_arguments = []
        for classifier_name in classifier_names:
            arguments = self.get_path_dict(kwargs_init[classifier_name])
            if hps_method == "Grid":
                multiview_arguments += [
                    self.gen_single_multiview_arg_dictionary(classifier_name,
                                                        arguments,
                                                        nb_class,
                                                        {"param_grid": hps_kwargs[
                                                            classifier_name]},
                                                        views_dictionary=views_dictionary)]
            elif hps_method == "Random":
                hps_kwargs = self.get_random_hps_args(hps_kwargs, classifier_name)
                multiview_arguments += [
                    self.gen_single_multiview_arg_dictionary(classifier_name,
                                                        arguments,
                                                        nb_class,
                                                        hps_kwargs,
                                                        views_dictionary=views_dictionary)]
            elif hps_method == "None":
                multiview_arguments += [
                    self.gen_single_multiview_arg_dictionary(classifier_name,
                                                        arguments,
                                                        nb_class,
                                                        hps_kwargs,
                                                        views_dictionary=views_dictionary)]
            else:
                raise ValueError('At the moment only "None",  "Random" or "Grid" '
                                 'are available as hyper-parameter search '
                                 'methods, sadly "{}" is not'.format(hps_method)
                                 )

        return multiview_arguments


    def init_monoview_exps(self, classifier_names,
                           views_dictionary, nb_class, kwargs_init, hps_method,
                           hps_kwargs):  # pragma: no cover
        r"""Used to add each monoview exeperience args to the list of monoview experiences args.

        First this function will check if the benchmark need mono- or/and multiview algorithms and adds to the right
        dictionary the asked algorithms. If none is asked by the user, all will be added.

        If the keyword `"Benchmark"` is used, all mono- and multiview algorithms will be added.

        Parameters
        ----------
        classifier_names : dictionary
            All types of monoview and multiview experiments that have to be benchmarked
        argument_dictionaries : dictionary
            Maps monoview and multiview experiments arguments.
        views_dictionary : dictionary
            Maps the view names to their index in the HDF5 dataset
        nb_class : integer
            Number of different labels in the classification

        Returns
        -------
        benchmark : Dictionary of dictionaries
            Dictionary resuming which mono- and multiview algorithms which will be used in the benchmark.
        """
        monoview_arguments = []
        for view_name, view_index in views_dictionary.items():
            for classifier_name in classifier_names:
                if hps_method == "Grid":
                    monoview_exp = self.gen_single_monoview_arg_dictionary(classifier_name=classifier_name,
                                                                   arguments=kwargs_init,
                                                                   view_index=view_index,
                                                                   view_name=view_name,
                                                                   hps_kwargs={"param_grid":
                                                                    hps_kwargs[
                                                                        classifier_name]})
                elif hps_method == "Random":
                    hps_kwargs = self.get_random_hps_args(hps_kwargs, classifier_name)
                    monoview_exp = self.gen_single_monoview_arg_dictionary(classifier_name=classifier_name,
                                                                   arguments=kwargs_init,
                                                                   view_index=view_index,
                                                                   view_name=view_name,
                                                                   hps_kwargs=hps_kwargs)
                elif hps_method == "None":
                    monoview_exp = self.gen_single_monoview_arg_dictionary(classifier_name=classifier_name,
                                                                   arguments=kwargs_init,
                                                                   view_index=view_index,
                                                                   view_name=view_name,
                                                                   hps_kwargs=hps_kwargs)

                else:
                    raise ValueError(
                        'At the moment only "None",  "Random" or "Grid" '
                        'are available as hyper-parameter search '
                        'methods, sadly "{}" is not'.format(hps_method)
                    )
                monoview_arguments.append(monoview_exp)
        return monoview_arguments

    def get_random_hps_args(self, hps_args, classifier_name):
        hps_dict = {}
        for key, value in hps_args.items():
            if key in ["n_iter", "equivalent_draws"]:
                hps_dict[key] = value
            if key == classifier_name:
                hps_dict["param_distributions"] = value
        return hps_dict

    def gen_single_monoview_arg_dictionary(self, classifier_name='', arguments={},
                                           view_index=0, view_name='',hps_kwargs={}):
        if classifier_name in arguments:
            classifier_config = dict((key, value) for key, value in arguments[
                classifier_name].items())
        else:
            classifier_config = {}
        plif = MonoViewExp(classifier_config=classifier_config,
                view_name= view_name, view_index= view_index,
                classifier_name=classifier_name,
                nb_class=self.nb_class,
                hps_kwargs=hps_kwargs,
                train_size=self.split,
                           database_name=self.name,
                           hps_type=self.hps_type, nb_cores=self.nb_cores,
                           labels_dictionary=self.labels_dictionary,
                           metrics=self.metrics)
        return plif


    def gen_single_multiview_arg_dictionary(self, classifier_name, arguments, nb_class,
                                            hps_kwargs, views_dictionary=None,):
        if classifier_name in arguments:
            classifier_config = dict((key, value) for key, value in arguments[
                classifier_name].items())
        else:
            classifier_config = {}
        return MultiViewExp(classifier_name=classifier_name,
                view_names=list(views_dictionary.keys()),
                view_indices=list(views_dictionary.values()),
                nb_class=nb_class,
                labels_dictionary=self.labels_dictionary,
                hps_kwargs=hps_kwargs,
                # classifier_name= self.extract_dict(arguments),
                           train_size=self.split,
                           classifier_config=classifier_config,
                           database_name=self.name,
                           hps_type=self.hps_type,
                           nb_cores=self.nb_cores,
                           metrics=self.metrics,

                           )


    def extract_dict(self, classifier_config):
        """Reverse function of get_path_dict"""
        extracted_dict = {}
        for key, value in classifier_config.items():
            extracted_dict = self.set_element(extracted_dict, key, value)
        return extracted_dict


    def set_element(self, dictionary, path, value):
        """Set value in dictionary at the location indicated by path"""
        existing_keys = path.split(".")[:-1]
        dict_state = dictionary
        for existing_key in existing_keys:
            if existing_key in dict_state:
                dict_state = dict_state[existing_key]
            else:
                dict_state[existing_key] = {}
                dict_state = dict_state[existing_key]
        dict_state[path.split(".")[-1]] = value
        return dictionary


    def get_path_dict(self, multiview_classifier_args):
        """This function is used to generate a dictionary with each key being
        the path to the value.
        If given {"key1":{"key1_1":value1}, "key2":value2}, it will return
        {"key1.key1_1":value1, "key2":value2}"""
        path_dict = dict(
            (key, value) for key, value in multiview_classifier_args.items())
        paths = self.is_dict_in(path_dict)
        while paths:
            for path in paths:
                for key, value in path_dict[path].items():
                    path_dict[".".join([path, key])] = value
                path_dict.pop(path)
            paths = self.is_dict_in(path_dict)
        return path_dict


    def is_dict_in(self, dictionary):
        """
        Returns True if any of the dictionary value is a dictionary itself.

        Parameters
        ----------
        dictionary

        Returns
        -------

        """
        paths = []
        for key, value in dictionary.items():
            if isinstance(value, dict):
                paths.append(key)
        return paths


    def init_kwargs(self, classifiers_names, framework="monoview"):
        r"""Used to init kwargs thanks to a function in each monoview classifier package.

        Parameters
        ----------
        args : parsed args objects
            All the args passed by the user.
        classifiers_names : list of strings
            List of the benchmarks's monoview classifiers names.

        Returns
        -------
        kwargs : Dictionary
            Dictionary resuming all the specific arguments for the benchmark, one dictionary for each classifier.

            For example, for Adaboost, the KWARGS will be `{"n_estimators":<value>, "base_estimator":<value>}`"""

        logging.info("Start:\t Initializing {} classifiers arguments".format(framework))
        kwargs = {}
        for classifiers_name in classifiers_names:
            try:
                if framework == "monoview":
                    getattr(monoview_classifiers, classifiers_name)
                else:
                    if classifiers_name.startswith('early_fusion_'):
                        getattr(monoview_classifiers, classifiers_name.split('early_fusion_')[1])
                    else:
                        getattr(multiview_classifiers, classifiers_name)
            except AttributeError:
                raise AttributeError(
                    classifiers_name + " is not implemented in monoview_classifiers, "
                                       "please specify the name of the file in monoview_classifiers")
            if classifiers_name in self.args:
                kwargs[classifiers_name] = self.args[classifiers_name]
            else:
                kwargs[classifiers_name] = {}
        logging.info("Done:\t Initializing {} classifiers arguments".format(framework))
        return kwargs

    def init_kwargs_func(self,):
        """
        Dispached the kwargs initialization to monoview and multiview and creates
        the kwargs variable

        Parameters
        ----------
        args : parsed args objects
            All the args passed by the user.

        benchmark : dict
            The name of the mono- and mutli-view classifiers to run in the benchmark

        Returns
        -------

        kwargs : dict
            The arguments for each mono- and multiview algorithms
        """
        if self.monoview_benchmark is not None:
            self.monoview_kwargs = self.init_kwargs(self.monoview_benchmark,
                                      framework="monoview")
        if self.multiview_benchmark is not None:
            self.multiview_kwargs = self.init_kwargs(self.multiview_benchmark,
                                       framework="multiview")


    def arange_metrics(self,):
        """Used to get the metrics list in the right order so that
        the first one is the principal metric specified in args

        Parameters
        ----------
        metrics : dict
            The metrics that will be used in the benchmark

        metric_princ : str
            The name of the metric that need to be used for the hyper-parameter
            optimization process

        Returns
        -------
        metrics : list of lists
            The metrics list, but arranged  so the first one is the principal one."""
        if self.metric_princ in self.metrics:
            self.metrics = dict(
                (key, value) if not key == self.metric_princ else (key + "*", value) for
                key, value in self.metrics.items())
        else:
            raise ValueError("{} not in metric pool ({})".format(self.metric_princ,
                                                                 self.metrics))







    def exec_benchmark(self, ):  # pragma: no cover
        r"""Used to execute the needed benchmark(s) on multicore or mono-core functions.

        Parameters
        ----------
        nb_cores : int
            Number of threads that the benchmarks can use.
        stats_iter : int
            Number of statistical iterations that have to be done.
        benchmark_arguments_dictionaries : list of dictionaries
            All the needed arguments for the benchmarks.
        classification_indices : list of lists of numpy.ndarray
            For each statistical iteration a couple of numpy.ndarrays is stored with the indices for the training set and
            the ones of the testing set.
        directories : list of strings
            List of the paths to the result directories for each statistical iteration.
        directory : string
            Path to the main results directory.
        multi_class_labels : ist of lists of numpy.ndarray
            For each label couple, for each statistical iteration a triplet of numpy.ndarrays is stored with the
            indices for the biclass training set, the ones for the biclass testing set and the ones for the
            multiclass testing set.
        metrics : list of lists
            metrics that will be used to evaluate the algorithms performance.
        labels_dictionary : dictionary
            Dictionary mapping labels indices to labels names.
        nb_labels : int
            Total number of different labels in the dataset.
        dataset_var : HDF5 dataset file
            The full dataset that wil be used by the benchmark.
        classifiers_names : list of strings
            List of the benchmarks's monoview classifiers names.
        rest_of_the_args :
            Just used for testing purposes


        Returns
        -------
        results : list of lists
            The results of the benchmark.
        """
        logging.info("Start:\t Executing all the needed benchmarks")
        self.results = []
        for iter_ind, bootstrap_iter in enumerate(self.benchmark_argument_dictionaries):
            bootstrap_iter.exec_one_benchmark_mono_core(self.dataset_var)
            analyze_iterations([bootstrap_iter], self.stats_iter,
                               self.metrics, sample_ids=self.dataset_var.sample_ids,
                               labels=self.dataset_var.get_labels(),
                               feature_ids=self.dataset_var.feature_ids,
                               view_names=self.dataset_var.view_names)
            self.results += [bootstrap_iter]
            self.benchmark_argument_dictionaries[iter_ind] = bootstrap_iter
        logging.info("Done:\t Executing all the needed benchmarks")

        # Do everything with flagging
        logging.info("Start:\t Analyzing predictions")
        results_mean_stds = analyze(self.results, self.stats_iter,
                                    self.benchmark_argument_dictionaries,
                                    self.metrics,
                                    self.res_dir,
                                    self.dataset_var.sample_ids,
                                    self.dataset_var.get_labels(),self.dataset_var.feature_ids,
                                    self.dataset_var.view_names)
        logging.info("Done:\t Analyzing predictions")
        return results_mean_stds


