import logging
import os
import os.path
import time

import h5py
import numpy as np

from .multiview_utils import MultiviewResult, MultiviewResultAnalyzer
from .. import multiview_classifiers
from .. import monoview_classifiers
from ..monoview.exec_classif_mono_view import MonoViewExp
from ..utils import hyper_parameter_search
from ..utils.multiclass import get_mc_estim
from ..utils.organization import secure_file_path

# Author-Info
__author__ = "Baptiste Bauvin"
__status__ = "Prototype"  # Production, Development, Prototype


class MultiViewExp(MonoViewExp):

    def __init__(self, classifier_name="decision_tree",
                 classifier_config={"depth":3}, view_names=[],
                 view_indices=[0], nb_class=2, hps_kwargs={}, train_size=0.8,
                 labels_dictionary={}, database_name="",
                 hps_type="Random", nb_cores=1, metrics={},
                 equivalent_draws=False):
        MonoViewExp.__init__(self, classifier_name=classifier_name,
                             classifier_config=classifier_config,
                             nb_class=nb_class,
                             hps_kwargs=hps_kwargs,
                             train_size=train_size,
                             labels_dictionary=labels_dictionary,
                             database_name=database_name,
                             hps_type=hps_type,
                             nb_cores=nb_cores, metrics=metrics)
        self.view_names = view_names
        self.view_indices = view_indices
        self.equivalent_draws = equivalent_draws

    def init_constants(self, dataset_var ):
        """
        Used to init the constants
        Parameters
        ----------
        kwargs :

        classification_indices :

        metrics :

        name :

        nb_cores : nint number of cares to execute

        k_folds :

        dataset_var :  {array-like} shape (n_samples, n_features)
            dataset variable

        Returns
        -------
        tuple of (classifier_name, t_start, views_indices,
                  classifier_config, views, learning_rate)
        """
        # learning_rate = len(self.split[0]) / float(
        #     (len(classification_indices[0]) + len(classification_indices[1])))
        t_start = time.time()
        logging.info("Info\t: Classification - Database : " + str(
            self.database_name) + " ; Views : " + ", ".join(self.view_names) +
                     " ; Algorithm : " + self.classifier_name + " ; Cores : " + str(
            self.nb_cores) + ", Train ratio : " + str(self.train_size) +
                     ", CV on " + str(self.k_folds.n_splits) + " folds")

        for view_index, view_name in zip(self.view_indices, self.view_names):
            logging.info("Info:\t Shape of " + str(view_name) + " :" + str(
                dataset_var.get_shape(view_index)))
        # labels = dataset_var.get_labels()
        self.directory = os.path.join(self.directory, self.classifier_name)
        self.base_file_name = self.classifier_name + "-" + dataset_var.get_name() + "-"
        self.output_file_name = os.path.join(self.directory, self.base_file_name)
        # return classifier_name, t_start, views_indices, \
        #        classifier_config, views, learning_rate, labels, output_file_name, \
        #        directory, base_file_name, metrics

    def save_results(self, string_analysis, images_analysis,
                     confusion_matrix):  # pragma: no cover
        """
        Save results in derectory

        Parameters
        ----------

        classifier : classifier class

        labels_dictionary : dict dictionary of labels

        string_analysis : str

        views :

        classifier_module : module of the classifier

        classification_kargs :

        directory : str directory

        learning_rate :

        name :

        images_analysis :

        """
        logging.info(string_analysis)
        secure_file_path(self.output_file_name)
        output_text_file = open(self.output_file_name + 'summary.txt', 'w',
                                encoding="utf-8")
        output_text_file.write(string_analysis)
        output_text_file.close()
        np.savetxt(self.output_file_name + "confusion_matrix.csv", confusion_matrix,
                   delimiter=',')

        if images_analysis is not None:
            for image_name in images_analysis.keys():
                if os.path.isfile(self.output_file_name + image_name + ".png"):
                    for i in range(1, 20):
                        test_file_name = self.output_file_name + image_name + "-" + str(
                            i) + ".png"
                        if not os.path.isfile(test_file_name):
                            images_analysis[image_name].savefig(test_file_name,
                                                                transparent=True)
                            break

                images_analysis[image_name].savefig(
                    self.output_file_name + image_name + '.png', transparent=True)

    def hdf5_to_monoview(self, dataset):
        """Here, we concatenate the views for the asked samples """
        monoview_data = np.concatenate(
            [dataset.get_v(view_idx)
             for index, view_idx
             in enumerate(self.view_indices)], axis=1)
        self.feature_ids = []
        for view_idx in self.view_indices:
            view_name = dataset.view_names[view_idx]
            self.feature_ids += [view_name+"-"+feat_id for feat_id in dataset.feature_ids[view_idx]]
        return monoview_data

    def exec(self, dataset_var, ):
        """Used to execute multiview classification and result analysis

        Parameters
        ----------

        directory : indicate the directory


        dataset_var :

        name

        classification_indices

        k_folds

        nb_cores

        database_type

        path

        labels_dictionary : dict dictionary of labels

        random_state : int seed, RandomState instance, or None (default=None)
            The seed of the pseudo random number multiview_generator to use when
            shuffling the data.

        labels

        hps_method

        metrics

        n_iter : int number of iterations

        kwargs

        Returns
        -------

        ``MultiviewResult``
        """
        t_start = time.monotonic()
        logging.info("Start:\t Initialize constants")
        self.init_constants(dataset_var)
        logging.info("Done:\t Initialize constants")

        extraction_time = time.time() - t_start
        logging.info(
            "Info:\t Extraction duration " + str(extraction_time) + "s")

        logging.info("Start:\t Getting train/test split")
        learning_indices, validation_indices = self.splits
        logging.info("Done:\t Getting train/test split")

        logging.info("Start:\t Getting classifiers modules")

        logging.info("Done:\t Getting classifiers modules")
        if self.classifier_name.startswith("early_fusion_"):
            mono_clf_name = self.classifier_name.split("early_fusion_")[1]
            classifier_module = getattr(monoview_classifiers, mono_clf_name)
            classifier_class_name = classifier_module.classifier_class_name
            X = self.hdf5_to_monoview(dataset_var)
            y = dataset_var.get_labels()
            train_pred, test_pred, X_train, y_train, y_test, \
            hps_duration, fit_duration, pred_duration = self.hp_train_test(X, y,
                                                                           classifier_class_name,
                                                                           classifier_module)
            self.classifier.short_name="EF_"+mono_clf_name
        else:
            self.feature_ids = dataset_var.feature_ids
            classifier_module = getattr(multiview_classifiers,
                                        self.classifier_name)
            classifier_class_name = classifier_module.classifier_class_name
            logging.info("Start:\t Optimizing hyperparameters")
            hps_beg = time.monotonic()

            if self.hps_type != "None":
                hps_method_class = getattr(hyper_parameter_search, self.hps_type)
                estimator = getattr(classifier_module, classifier_class_name)(
                    random_state=self.random_state,
                    **self.classifier_config)
                estimator = get_mc_estim(estimator, self.random_state,
                                         multiview=True,
                                         y=dataset_var.get_labels()[
                                             learning_indices])
                hps = hps_method_class(estimator, scoring=self.metrics, cv=self.k_folds,
                                       random_state=self.random_state,
                                       framework="multiview",
                                       n_jobs=self.nb_cores,
                                       learning_indices=learning_indices,
                                       view_indices=self.view_indices,
                                       **self.hps_kwargs)
                hps.fit(dataset_var, dataset_var.get_labels(), )
                classifier_config = hps.get_best_params()
                hps.gen_report(self.output_file_name)
            hps_duration = time.monotonic() - hps_beg
            self.classifier = get_mc_estim(
                getattr(classifier_module, classifier_class_name)(
                    random_state=self.random_state,
                    **self.classifier_config),
                self.random_state, multiview=True,
                y=dataset_var.get_labels())
            logging.info("Done:\t Optimizing hyperparameters")
            logging.info("Start:\t Fitting classifier")
            fit_beg = time.monotonic()

            self.classifier.fit(dataset_var, dataset_var.get_labels(),
                           train_indices=learning_indices,
                           view_indices=self.view_indices)

            fit_duration = time.monotonic() - fit_beg
            logging.info("Done:\t Fitting classifier")

            logging.info("Start:\t Predicting")
            train_pred = self.classifier.predict(dataset_var,
                                            sample_indices=learning_indices,
                                            view_indices=self.view_indices)
            pred_beg = time.monotonic()
            test_pred = self.classifier.predict(dataset_var,
                                           sample_indices=validation_indices,
                                           view_indices=self.view_indices)
            pred_duration = time.monotonic() - pred_beg
        full_pred = np.zeros(dataset_var.get_labels().shape, dtype=int) - 100
        full_pred[learning_indices] = train_pred
        full_pred[validation_indices] = test_pred
        logging.info("Done:\t Pertidcting")

        whole_duration = time.time() - t_start
        logging.info(
            "Info:\t Classification duration " + str(extraction_time) + "s")

        logging.info("Start:\t Result Analysis for " + self.classifier_name)
        times = (extraction_time, whole_duration)
        if "n_iter" in self.hps_kwargs:
            self.n_iter_hps = self.hps_kwargs["n_iter"]
        else:
            self.n_iter_hps = 0
        result_analyzer = MultiviewResultAnalyzer(view_names=self.view_names,
                                                  classifier=self.classifier,
                                                  classification_indices=self.splits,
                                                  k_folds=self.k_folds,
                                                  hps_method=self.hps_type,
                                                  metrics_dict=self.metrics,
                                                  n_iter=self.n_iter_hps,
                                                  class_label_names=[ self.labels_dictionary[ind]
                                                      for ind in range(len(self.labels_dictionary))],
                                                  pred=full_pred,
                                                  directory=self.directory,
                                                  base_file_name=self.base_file_name,
                                                  labels=dataset_var.get_labels(),
                                                  database_name=dataset_var.get_name(),
                                                  nb_cores=self.nb_cores,
                                                  duration=whole_duration,
                                                  feature_ids=self.feature_ids)
        string_analysis, images_analysis, metrics_scores, class_metrics_scores, \
        confusion_matrix = result_analyzer.analyze()
        logging.info("Done:\t Result Analysis for " + self.classifier_name)

        logging.info("Start:\t Saving preds")
        self.save_results(string_analysis, images_analysis, confusion_matrix)
        logging.info("Start:\t Saving preds")

        return MultiviewResult(self.classifier_name, self.classifier_config, metrics_scores,
                               full_pred, hps_duration, fit_duration,
                               pred_duration, class_metrics_scores, self.classifier)





# def exec_multiview_multicore(directory, core_index, name, learning_rate,
#                              nb_folds,
#                              database_type, path, labels_dictionary,
#                              random_state, labels,
#                              hyper_param_search=False, nb_cores=1, metrics=None,
#                              n_iter=30, **arguments):  # pragma: no cover
#     """
#     execute multiview process on
#
#     Parameters
#     ----------
#
#     directory : indicate the directory
#
#     core_index :
#
#     name : name of the data file to perform
#
#     learning_rate :
#
#     nb_folds :
#
#     database_type :
#
#     path : path to the data name
#
#     labels_dictionary
#
#     random_state : int seed, RandomState instance, or None (default=None)
#         The seed of the pseudo random number multiview_generator to use when
#         shuffling the data.
#
#     labels :
#
#     hyper_param_search :
#
#     nb_cores : in number of cores
#
#     metrics : metric to use
#
#     n_iter : int number of iterations
#
#     arguments : others arguments
#
#     Returns
#     -------
#     exec_multiview on directory, dataset_var, name, learning_rate, nb_folds, 1,
#         database_type, path, labels_dictionary,
#         random_state, labels,
#         hyper_param_search=hyper_param_search, metrics=metrics,
#         n_iter=n_iter, **arguments
#     """
#     """Used to load an HDF5 dataset_var for each parallel job and execute multiview classification"""
#     dataset_var = h5py.File(path + name + str(core_index) + ".hdf5", "r")
#     return exec_multiview(directory, dataset_var, name, learning_rate, nb_folds,
#                           1,
#                           database_type, path, labels_dictionary,
#                           random_state, labels,
#                           hps_method=hyper_param_search,
#                           metrics=metrics,
#                           n_iter=n_iter, **arguments)



