from sklearn.tree import DecisionTreeClassifier
from sklearn.base import BaseEstimator

import numpy as np
from multimodal.boosting.combo import MuComboClassifier
from ..multiview.multiview_utils import BaseMultiviewClassifier
from ..utils.hyper_parameter_search import CustomRandint
from ..utils.dataset import get_samples_views_indices
from ..utils.base import base_boosting_estimators

classifier_class_name = "MuCombo"


class MuCombo(BaseMultiviewClassifier, MuComboClassifier):

    def __init__(self, base_estimator=None,
                 n_estimators=50,
                 random_state=None,**kwargs):
        BaseMultiviewClassifier.__init__(self, random_state)
        base_estimator = self.set_base_estim_from_dict(base_estimator, **kwargs)
        MuComboClassifier.__init__(self, base_estimator=base_estimator,
                                    n_estimators=n_estimators,
                                    random_state=random_state,)
        self.param_names = ["base_estimator", "n_estimators", "random_state",]
        self.distribs = [base_boosting_estimators,
                         CustomRandint(5,200), [random_state],]

    def fit(self, X, y, train_indices=None, view_indices=None):
        train_indices, view_indices = get_samples_views_indices(X,
                                                                 train_indices,
                                                                 view_indices)
        self.used_views = view_indices
        numpy_X, view_limits = X.to_numpy_array(sample_indices=train_indices,
                                                view_indices=view_indices)
        self.view_shapes = [view_limits[ind + 1] - view_limits[ind]
                            for ind in range(len(self.used_views))]
        return MuComboClassifier.fit(self, numpy_X, y[train_indices],
                                                view_limits)

    def set_params(self, base_estimator=None, **params):
        """
        Sets the base estimator from a dict.
        :param base_estimator:
        :param params:
        :return:
        """
        if base_estimator is None:
            self.base_estimator = DecisionTreeClassifier()
        elif type(base_estimator) is list:
            if type(base_estimator[0]) is dict:
                self.base_estimator = [self.set_base_estim_from_dict(estim) for estim in base_estimator]
            elif isinstance(base_estimator[0], BaseEstimator):
                self.base_estimator = base_estimator
            else:
                raise ValueError("base_estimator should ba a list of dict or a sklearn classifier list")
        elif isinstance(base_estimator, dict):
            self.base_estimator = self.set_base_estim_from_dict(base_estimator)
            MuComboClassifier.set_params(self, **params)
        else:
            MuComboClassifier.set_params(self, base_estimator=base_estimator, **params)

    def predict(self, X, sample_indices=None, view_indices=None):
        sample_indices, view_indices = get_samples_views_indices(X,
                                                                 sample_indices,
                                                                 view_indices)
        self._check_views(view_indices)
        numpy_X, view_limits = X.to_numpy_array(sample_indices=sample_indices,
                                                view_indices=view_indices)
        return MuComboClassifier.predict(self, numpy_X)

    def accepts_multi_class(self, random_state, n_samples=10, dim=2,
                            n_classes=3, n_views=2):
        return True

    def get_interpretation(self, directory, base_file_name, labels,
                           multiclass=False):
        self.feature_importances_ = [np.zeros(view_shape)
                                     for view_shape in self.view_shapes]
        # for best_view, estimator_weight, estimator in zip(self.best_views_, self.estimator_weights_, self.estimators_):
        #     self.view_importances[best_view] += estimator_weight
        #     if hasattr(estimator, "feature_importances_"):
        #         self.feature_importances_[best_view] += estimator.feature_importances_
        # importances_sum = sum([np.sum(feature_importances)
        #                        for feature_importances
        #                        in self.feature_importances_])
        # self.feature_importances_ = [feature_importances/importances_sum
        #                              for feature_importances
        #                              in self.feature_importances_]
        return ""
